#include "mns.h"

struct iplist *
getlocalips (void)
{
  struct iplist *iplhead, *iplwalk, *iplold = NULL;
  struct ifi_info *ifi, *ifihead;

  iplhead = iplwalk = calloc (1, sizeof (struct iplist));
#define DOALIASES 0
  ifihead = ifi = get_ifi_info (AF_INET, DOALIASES);

  printf ("Interfaces Detected: \n");

  for (; ifi != NULL; ifi = ifi->ifi_next)
    {
      if (memcmp (ifi->ifi_name, "lo", 2))
	{
	  printf ("%8s ", ifi->ifi_name);
#define SAifi (SAin *)ifi
	  if (SAifi->ifi_addr != NULL)
	    {
	      printf ("IP: %s\n", inet_ntoa((SAifi->ifi_addr)->sin_addr));
	      iplold = iplwalk;
	      iplwalk->ip = (SAifi->ifi_addr)->sin_addr;
	      iplwalk->next = calloc (1, sizeof (struct iplist));
	      iplwalk = iplwalk->next;
	    }
	}
    }

  if (iplold == NULL)
    {
      free (iplhead);
      iplhead = NULL;
    }
  else
    {
      free (iplold->next);
      iplold->next = NULL;
    }

  free_ifi_info (ifihead);
  return iplhead;
}

struct ifi_info *
get_ifi_info (int family, int doaliases)
{
  struct ifi_info *ifi, *ifihead, **ifipnext;
  int sockfd, len, lastlen, flags, myflags;
  char *ptr, *buf, lastname[IFNAMSIZ], *cptr;
  struct ifconf ifc;
  struct ifreq *ifr, ifrcopy;
  SAin *sinptr;

  sockfd = socket (AF_INET, SOCK_DGRAM, 0);

  lastlen = 0;
  len = 100 * sizeof (struct ifreq);
  for (;;)
    {

      buf = malloc (len);
      ifc.ifc_len = len;
      ifc.ifc_buf = buf;
      if (ioctl (sockfd, SIOCGIFCONF, &ifc) < 0)
	{
	  if (errno != EINVAL || lastlen != 0)
	    {
	      fprintf (stderr, "ioctl error\n");
	      exit (EXIT_FAILURE);
	    }
	}
      else
	{
	  if (ifc.ifc_len == lastlen)
	    break;
	  lastlen = ifc.ifc_len;
	}

      len += 10 * sizeof (struct ifreq);	/* increment */
      free (buf);
    }

  ifihead = NULL;
  ifipnext = &ifihead;
  lastname[0] = 0;
/* end get_ifi_info1 */

/* include get_ifi_info2 */
  for (ptr = buf; ptr < buf + ifc.ifc_len;)
    {
      ifr = (struct ifreq *) ptr;

#ifdef	HAVE_SOCKADDR_SA_LEN
      len = max (sizeof (SA), ifr->ifr_addr.sa_len);
#else
      switch (ifr->ifr_addr.sa_family)
	{
	case AF_INET:
	default:
	  len = sizeof (SA);
	  break;
	}
#endif /* HAVE_SOCKADDR_SA_LEN */
      /* for next one in buffer */
      ptr += sizeof (ifr->ifr_name) + len;

      /* ignore if not desired address family */
      if (ifr->ifr_addr.sa_family != family)
	continue;

      myflags = 0;
      if ((cptr = strchr (ifr->ifr_name, ':')) != NULL)
	*cptr = 0;		/* replace colon will null */
      if (strncmp (lastname, ifr->ifr_name, IFNAMSIZ) == 0)
	{
	  /* already processed this interface */
	  if (doaliases == 0)
	    continue;
	  myflags = IFI_ALIAS;
	}

      memcpy (lastname, ifr->ifr_name, IFNAMSIZ);

      ifrcopy = *ifr;
      ioctl (sockfd, SIOCGIFFLAGS, &ifrcopy);
      flags = ifrcopy.ifr_flags;
      /* ignore if interface not up */
      if ((flags & IFF_UP) == 0)
	continue;

      ifi = calloc (1, sizeof (struct ifi_info));
      *ifipnext = ifi;		/* prev points to this new one */
      ifipnext = &ifi->ifi_next;	/* pointer to next one goes here */

      ifi->ifi_flags = flags;	/* IFF_xxx values */
      ifi->ifi_myflags = myflags;	/* IFI_xxx values */
      memcpy (ifi->ifi_name, ifr->ifr_name, IFI_NAME);
      ifi->ifi_name[IFI_NAME - 1] = '\0';
/* end get_ifi_info2 */
/* include get_ifi_info3 */
      switch (ifr->ifr_addr.sa_family)
	{
	case AF_INET:
	  sinptr = (SAin *) & ifr->ifr_addr;
	  if (ifi->ifi_addr == NULL)
	    {
	      ifi->ifi_addr = calloc (1, sizeof (SAin));
	      memcpy (ifi->ifi_addr, sinptr, sizeof (SAin));
/*
   #ifdef       SIOCGIFBRDADDR
   if (flags & IFF_BROADCAST) {
   ioctl(sockfd, SIOCGIFBRDADDR, &ifrcopy);
   sinptr = (SAin *)&ifrcopy.ifr_broadaddr;
   ifi->ifi_brdaddr = calloc(1, sizeof(SAin));
   memcpy(ifi->ifi_brdaddr, sinptr, sizeof(SAin));
   }
   #endif
 */
/*
   #ifdef       SIOCGIFDSTADDR
   if (flags & IFF_POINTOPOINT) {
   ioctl(sockfd, SIOCGIFDSTADDR, &ifrcopy);
   sinptr = (SAin *) &ifrcopy.ifr_dstaddr;
   ifi->ifi_dstaddr = calloc(1, sizeof(SAin));
   memcpy(ifi->ifi_dstaddr, sinptr, sizeof(SAin));
   }
   #endif
 */
	    }
	  break;

	default:
	  break;
	}
    }
  free (buf);
  return (ifihead);		/* pointer to first structure in linked list */
}
/* end get_ifi_info3 */

/* include free_ifi_info */
void 
free_ifi_info (struct ifi_info *ifihead)
{
  struct ifi_info *ifi, *ifinext;

  for (ifi = ifihead; ifi != NULL; ifi = ifinext)
    {
      if (ifi->ifi_addr != NULL)
	free (ifi->ifi_addr);

/*      if (ifi->ifi_brdaddr != NULL) free(ifi->ifi_brdaddr); */
/*      if (ifi->ifi_dstaddr != NULL) free(ifi->ifi_dstaddr); */

      ifinext = ifi->ifi_next;	/* can't fetch ifi_next after free() */
      free (ifi);		/* the ifi_info{} itself */
    }
}
/* end free_ifi_info */
