#define VERSION "mns-v.68beta"

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <rpc/pmap_clnt.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <fcntl.h>
#include <time.h>

#define OSCFGFILE "fingerprints.conf"

#define VULNQPOP (1     )
#define VULNIMAP (1 << 1)
#define VULNNLPS (1 << 2)
#define VULNCGI  (1 << 3)
#define VULNRPC  (1 << 4)
#define VULNFTP  (1 << 5)
#define RESOLVE  (1 << 6)

#define ERROR -1
#define TIMEOUT 3

#define SA   struct sockaddr
#define SAin struct sockaddr_in

#define IPV4_ADDRLEN 16

#define stralloc(a,b)\
 {a->string = malloc(strlen(b) + 1);\
 strcpy((a)->string, (b));\
 (a)->next = calloc(sizeof(struct node),1);\
 a = (a)->next;}

/* TCP flags, even though we can use the ones in ip_tcp.h */
#define FIN 0x01
#define SYN 0x02
#define RST 0x04
#define PSH 0x08
#define ACK 0x10
#define URG 0x20
#define XXX 0x40
#define YYY 0x80

#define	IFI_NAME 16 /* same as IFNAMSIZ in <net/if.h> */
#define	IFI_HADDR 8 /* allow for 64-bit EUI-64 in future */
#define	IFI_ALIAS 1 /* ifi_addr is an alias */

struct ifi_info {
  char    ifi_name[IFI_NAME];	/* interface name, null terminated */
  u_char  ifi_haddr[IFI_HADDR];	/* hardware address */
  u_short ifi_hlen;		/* #bytes in hardware address: 0, 6, 8 */
  short   ifi_flags;		/* IFF_xxx constants from <net/if.h> */
  short   ifi_myflags;		/* our own IFI_xxx flags */
  struct sockaddr  *ifi_addr;	/* primary address */
/*  struct sockaddr  *ifi_brdaddr;*/ /* broadcast address */
/*  struct sockaddr  *ifi_dstaddr;*/ /* destination address */
  struct ifi_info  *ifi_next;	/* next of these structures */
};

struct node {
   char *string;
   struct node *next;
};

struct iplist {
   struct in_addr ip;
/*   struct in_addr mask; */
   struct iplist *next;
};

typedef struct {
   struct in_addr sip;
   struct in_addr dip;
   unsigned char zero;
   unsigned char proto;
   u_short tcplen;
} tcpsrec;

struct fingerprint {
   u_short set:2,
           seq:2,
           ackset:3,
           urg:1;
   u_long ack;
   u_short win;
   u_short flag;
};

typedef struct {
   SAin from;
   SAin dest;
   u_short sport;
   u_short dport;
   u_long seq;
   u_long ack;
} hdrinfo;

extern struct iplist *localiplist;

/* ipv4.c */
extern const char *ipv4_ntop(int, const void *, char *, size_t);
extern const char *ipv4_showname(int, const void *, char *, size_t);
extern int ipv4_rslvhost(char *, SAin *);
extern u_short in_cksum(u_short *, int);
extern u_short tcp_cksum(char *, int, struct in_addr, struct in_addr);
extern int hamming(u_long);

/* ifinfo.c */
extern struct ifi_info *get_ifi_info(int, int);
extern struct iplist *getlocalips(void);
extern void free_ifi_info(struct ifi_info *);

/* osdetect.c */
extern struct node *stealthscan(struct in_addr, struct node *, int **);
extern struct node *checkos(struct in_addr, struct node *, int *);

/* scan.c */
extern struct node *rpc(struct in_addr, struct node *);
extern struct node *qpop(struct in_addr, struct node *);
extern struct node *imap(struct in_addr, struct node *);
extern struct node *nlps(struct in_addr, struct node *);
extern struct node *cgi(struct in_addr, struct node *);
extern struct node *ftp(struct in_addr, struct node *);
extern void sighandler(int);

