#include "mns.h"

#define SERVICES 23
#define ACK_RAND 200
#define PACKETS 8
#define PORT_CLOSED 1
#define ip(a) (a).sin_addr.s_addr

void rawsockets(int *, int *);
void sendtcp(int, hdrinfo *, u_short);
char *checkosmain(int, int, SAin, SAin, short);
u_long gettcp(int, u_long, struct tcphdr *);
void synscan(int, int, int *, hdrinfo);
void finscan(int, int, int *, hdrinfo);

struct servicetype {
   char *name;
   int port;
};

static struct servicetype service[SERVICES] = {
   {     "echo", 7   }, {    "daytime", 13  }, {     "netstat", 15  },
   {  "chargen", 19  }, {        "ftp", 21  }, {         "ssh", 22  },
   {   "telnet", 23  }, {       "smtp", 25  }, {        "time", 37  },
   {   "domain", 53  }, {     "finger", 79  }, {         "www", 80  },
   { "kerberos", 88  }, {       "pop3", 110 }, {      "sunrcp", 111 },
   {     "auth", 113 }, { "netbios-ns", 137 }, { "netbios-ssn", 139 },
   {     "nmap", 143 }, {      "login", 513 }, {       "shell", 514 },
   {  "printer", 515 }, {     "listen",2766 }
};

struct node *stealthscan(struct in_addr hostip,
                         struct node *walk,
                         int **plist) {
   hdrinfo targetinfo;
   struct tcphdr itcp;
   struct iplist *iplwalk;
   struct in_addr interface;
   u_long ifip;
   SAin from, dest;
   int i, j, k;
   int sd_send, sd_recv;
   int pschkd[20] = { 0 };
   long timeout;
   long ack_diff = 0;
   char buffer[80];
   
   srand(time(NULL));

   bzero(&from, sizeof(SAin));
   bzero(&dest, sizeof(SAin));
   
   rawsockets(&sd_send, &sd_recv);

   interface.s_addr = 0;
   dest.sin_addr = hostip;
   targetinfo.sport = (rand() % 10000) + 5000;
   targetinfo.dest = dest;
   targetinfo.ack = 0;

   /* check if FIN scanning works on target host by the
    * single packet windows 9x/nt os check */

   targetinfo.seq = rand();
   for(iplwalk = localiplist; iplwalk != NULL; iplwalk = iplwalk->next) {
      from.sin_addr = iplwalk->ip;
      targetinfo.from = from;
      /* send to 2 random ports per interface incase one actually ends up
       * being open; for this check, the port must be closed */
      targetinfo.dport = (rand() % 1000) + 1000;
      sendtcp(sd_send, &targetinfo, SYN | FIN);
      
      targetinfo.dport = (rand() % 1000) + 1000;
      sendtcp(sd_send, &targetinfo, SYN | FIN);
   }

   timeout = time(NULL) + TIMEOUT;

   while(timeout > time(NULL)) {
      if((ifip = gettcp(sd_recv, ip(targetinfo.dest), &itcp))) {
         if(ntohl(itcp.th_ack)) {
            interface.s_addr = ifip;
            ack_diff = ntohl(itcp.th_ack) - targetinfo.seq;
            break;
         }
      }
   }
   
   usleep(300);

   from.sin_addr = interface;
   targetinfo.from = from;

   if(!interface.s_addr) {
      stralloc(walk, "\n   - target machine failed to respond");
      close(sd_send);
      close(sd_recv);
      *plist = NULL;
      return walk;
   }
   
   if(ack_diff == 2) {
      stralloc(walk, "\n   - non-standard response, reverting to SYN scan");
      stralloc(walk, "\n   - target likely to be a Windows based machine");
      synscan(sd_send, sd_recv, pschkd, targetinfo);
   } else finscan(sd_send, sd_recv, pschkd, targetinfo);

   for(i = j = 0; i < SERVICES; i++)
      if(!pschkd[i])
         j++;

   if(j) {
      stralloc(walk, "\n   TCP ports possibly open ->");

      *plist = malloc((j + 1) * sizeof(int) + sizeof(struct in_addr));
      (*plist)[0] = j;
      
      *(struct in_addr *)&(*plist)[j + 1] = interface;
      
      for(i = k = 0; i < SERVICES; i++) {
         if(!pschkd[i]) {
            k++;
            (*plist)[k] = service[i].port;
            sprintf(buffer, "\n      %3i   tcp\t\t%s",
                    service[i].port, service[i].name);
            stralloc(walk, buffer);
         }
      }
   } else *plist = NULL;
      
  close (sd_send);
  close (sd_recv);

  return walk;
}

struct node *checkos(struct in_addr hostip, struct node *walk, int *plist) {
  SAin dest, from;
  int sd_send, sd_recv;
  char *buffer = NULL;
  int i, pcount;
  char ham_asc[2];

  if(plist != NULL) pcount = plist[0] + 1;
  else return walk;

  rawsockets(&sd_send, &sd_recv);

  bzero (&dest, sizeof (SAin));
  bzero (&from, sizeof (SAin));

  dest.sin_addr = hostip;
  from.sin_addr = *(struct in_addr *)&plist[pcount];
  
  for (i = 1; i < pcount; i++) {
     buffer = checkosmain(sd_send, sd_recv, from, dest, plist[i]);
     if (buffer != NULL) {
        ham_asc[0] = buffer[0];
        ham_asc[1] = '\0';
        stralloc (walk, "\n   TCP Sequence Prediction Difficulty ->");
        stralloc (walk, ham_asc);
        switch (ham_asc[0]) {
           case '0': stralloc (walk, "( indeterminate )"); break;
           case '1': stralloc (walk, "( easy )"); break;
           case '2': stralloc (walk, "( easy )"); break;
           case '3': stralloc (walk, "( moderate )"); break;
           case '4': stralloc (walk, "( moderate )"); break;
           case '5': stralloc (walk, "( hard )"); break;
           case '6': stralloc (walk, "( infeaseable )"); break;
           case '7': stralloc (walk, "( impossible )"); break;
           case '8': stralloc (walk, "( impossible )"); break;
           default: stralloc (walk, "( hamming error )"); break;
        }
        stralloc (walk, "\n   Operating System ->");
        stralloc (walk, &buffer[1]);
        break;
     }
     if (buffer != NULL)
        break;
  }

  if (buffer != NULL)
     free (buffer);
  close (sd_send);
  close (sd_recv);

  return walk;
}

char *checkosmain(int sd_send, int sd_recv,
 	          SAin from, SAin dest, short dport) {
   FILE *fs_cfg;
   char *osname;
   struct fingerprint pak[PACKETS];
   struct tcphdr tcp;
   hdrinfo targetinfo;
   u_short start, s;
   u_long myseq[PACKETS];
   u_long seq_mask = 0, last_seq = 0;
   long timeout;
   char line[60];
   int n, i, ham;

   for (n = 0; n < PACKETS; n++)
      bzero(&pak[n], sizeof(struct fingerprint));

   srand(time(NULL) & 0x0000ffff);
   start = s = (u_short)getpid() + 5000 + (rand() % 5000);
   targetinfo.ack = 0;
   targetinfo.from = from;
   targetinfo.dest = dest;
   targetinfo.dport = dport;

#define sendtcppacket(a, b)\
   {targetinfo.seq = myseq[a] = rand();\
   targetinfo.sport = s + a;\
   sendtcp(sd_send, &targetinfo, b);\
   usleep(10);}

  sendtcppacket (0, SYN);
  sendtcppacket (1, SYN | ACK);
  sendtcppacket (2, FIN);
  sendtcppacket (3, FIN | ACK);
  sendtcppacket (4, SYN | FIN);
  sendtcppacket (5, PSH);
  sendtcppacket (6, SYN | XXX | YYY);
  sendtcppacket (7, SYN); /* this extra syn is used for hamming distances
                           * an even number of responses are needed */
  timeout = time (NULL) + TIMEOUT;
  n = i = 0;

  do {
     if (gettcp(sd_recv, ip(targetinfo.dest), &tcp)) {
        if(htons(targetinfo.dport) == tcp.th_sport) {
           if(ntohs(tcp.th_dport) - start < PACKETS) {
              n = ntohs(tcp.th_dport) - start;

              if((n < 0) || (n >= PACKETS)) continue;
              if(pak[n].set) continue;
              
	      if(!n && !ntohs(tcp.th_win)) return NULL;
              
              pak[n].seq = tcp.th_seq != 0;
              
	      if(pak[n].seq) {
                 seq_mask ^= tcp.th_seq;
                 last_seq = tcp.th_seq;
                 i++;
              }
              
              pak[n].ack = ntohl(tcp.th_ack);
              pak[n].ackset = tcp.th_ack != 0;
              if(pak[n].ackset && (myseq[n] > pak[n].ack))
                 pak[n].ackset = 2;
              if(pak[n].ackset) pak[n].ack = (myseq[n] < pak[n].ack) ?
                                                   (pak[n].ack - myseq[n]) :
                                                   (myseq[n] - pak[n].ack);
              if(pak[n].ack > ACK_RAND) {
                 pak[n].ackset = 3;
                 pak[n].ack = 0;
              }
              
              pak[n].win = ntohs(tcp.th_win);
              pak[n].flag = tcp.th_flags;
              pak[n].set = 1;
              pak[n].urg = tcp.th_urp != 0;
           }
        }
     }
     usleep(10);
  } while ((timeout > time (NULL)) && (n != PACKETS - 1));

/* calculate hamming distance */
  if(i % 2) seq_mask ^= last_seq;
  ham = hamming(seq_mask);
  if(ham > 8) ham = 16 - ham;

  /* parse fingerprints */
  if ((fs_cfg = fopen(OSCFGFILE, "r")) != NULL) {
     struct fingerprint parse;
     int flag1 = 0, found = 0, total = 0;
     char *token;

     osname = malloc(60);

     while (fgets(line, 60, fs_cfg)) {
        if(line[0] == '#') continue; /* COMMENT LINE */
        if(line[0] == '\n') { /* BLANK LINE */
           if (flag1 && (found == total)) {
              fclose(fs_cfg);
              osname[0] = (char)ham + '0';
              return osname;
           };
        }
        if(line[0] == '*') { /* OS FINGERPRINT HEADER */
           if (strlen(line) > 3) {
              strcpy(osname, &line[1]);
              osname[strlen(osname) - 1] = '\0';
           }
           flag1 = 1;
           total = found = 0;
           continue;
        }
           
        /* line number */
        token = strtok(line, " ");
        if(isdigit(token[0])) n = atoi(token);
	else {
           flag1 = total = found = 0;
	   continue;
        }

        /*-- seq --*/
	if((token = strtok(NULL, " ")) != NULL) parse.seq = atoi(token);
        parse.set = isdigit(token[0]) ? 1 : 0;
        
        /*-- ack --*/
	if((token = strtok(NULL, " ")) != NULL) {
           switch (token[0]) {
              case '+': parse.ackset = 1; parse.ack = atol(&token[1]); break;
              case '-': parse.ackset = 2; parse.ack = atol(&token[1]); break;
              case 'R': parse.ackset = 3; parse.ack = 0; break;
              case 'N':
              default: parse.ackset = 0; parse.ack = 0; break;
           }
        }

        /*-- win --*/
	if ((token = strtok(NULL, " ")) != NULL)
           parse.win = strtol(token, NULL, 16);

        /*-- flags --*/
        if((token = strtok (NULL, " \n")) != NULL) {
           parse.flag = 0;
           if (strchr(token, 'S')) parse.flag |= SYN;
           if (strchr(token, 'R')) parse.flag |= RST;
           if (strchr(token, 'A')) parse.flag |= ACK;
	   if (strchr(token, 'F')) parse.flag |= FIN;
           if (strchr(token, 'P')) parse.flag |= PSH;
           if (strchr(token, 'X')) parse.flag |= XXX;
           if (strchr(token, 'Y')) parse.flag |= YYY;
           if (strchr(token, 'U')) parse.urg = 1;
           else parse.urg = 0;
        }
        
        total++;
        
        /* test for no response */
        if (!pak[n].set) {
           if (parse.set) total = found = flag1 = 0;
           else found++;
           continue;
        }
        
        if ((parse.ackset == pak[n].ackset) &&
            (parse.ack == pak[n].ack) &&
            (parse.seq == pak[n].seq) &&
            (parse.win == pak[n].win) &&
            (parse.flag == pak[n].flag) &&
            (parse.urg == pak[n].urg))
        {
           found++;
           continue;
        } else {
           total = found = flag1 = 0;
           continue;
        }
     }

     fseek (fs_cfg, 0L, SEEK_END);
     fclose (fs_cfg);
     if (osname != NULL)
        free(osname);

     return NULL;
  }

  fprintf (stderr, "ERROR: cannot read from %s\n", OSCFGFILE);
  return NULL;
}

/* Sends a TCP packet */
void sendtcp(int sd_send, hdrinfo *routing, u_short flags) {
  struct tcphdr tcp;
  struct ip ip;
  static char pkt[8192];

/*-- IP HDR --*/
  ip.ip_hl = 5;
  ip.ip_v = 4;
  ip.ip_tos = 0;
  ip.ip_len = htons(40);
  ip.ip_id = htons(rand() % 0xffff);
  ip.ip_off = 0;
  ip.ip_ttl = 200 + (rand() % 56);
  ip.ip_p = IPPROTO_TCP;
  ip.ip_src = routing->from.sin_addr;
  ip.ip_dst = routing->dest.sin_addr;
#ifdef HAVE_STRUCT_IP_CSUM
#define ip_sum ip_csum
#endif
  ip.ip_sum = 0;
  ip.ip_sum = in_cksum((u_short *)&ip, sizeof(struct ip));

/*-- TCP HDR --*/
  tcp.th_sport = htons(routing->sport);
  tcp.th_dport = htons(routing->dport);
  tcp.th_seq = htonl(routing->seq);
  tcp.th_ack = 0;
#ifdef X2_OFF
  tcp.th_x2_off = 0x50;
#else
  tcp.th_x2 = 0;
  tcp.th_off = 5;
#endif
  tcp.th_flags = flags;
  tcp.th_win = htons(0x1000 | (rand() % 0x1000));
  tcp.th_urp = 0;
  tcp.th_sum = 0;

/*-- TCP Checksum --*/
#ifdef SOLARIS_CKSUM_BUG
  tcp.th_sum = sizeof(struct tcphdr);
#else
  tcp.th_sum = tcp_cksum((char *)&tcp,
			 sizeof (struct tcphdr),
			 routing->from.sin_addr,
			 routing->dest.sin_addr);
#endif /* SOLARIS_CKSUM_BUG */

  memcpy (pkt, &ip, sizeof(ip));
  memcpy (pkt + sizeof(ip), &tcp, sizeof(tcp));

  if (sendto(sd_send, pkt, 40, 0, (SA *)&routing->dest, sizeof(SA)) < 0)
     perror("sending message");

}

#define PACKETLEN 0xffffL
#define tcphd ((struct tcphdr *)&packet[20])
#define iphd  ((struct ip *)packet)

u_long gettcp(int sd_recv, u_long targetip, struct tcphdr *dtcp) {
   char packet[PACKETLEN];
  
   if(read(sd_recv, packet, PACKETLEN) < 0) return 0;
   if(targetip != iphd->ip_src.s_addr) return 0;
   
   memcpy(dtcp, tcphd, sizeof(struct tcphdr));
   
   return iphd->ip_dst.s_addr;
}

void rawsockets(int *sd_send, int *sd_recv) {
   int on = 1;
   int rsflags;

/*-- SEND RAW socket --*/
   if ((*sd_send = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
      perror ("opening raw send socket");
      exit (EXIT_FAILURE);
   }
   if (setsockopt(*sd_send, IPPROTO_IP, IP_HDRINCL, &on, sizeof(on)) < 0) {
      perror("setting option IP_HDRINCL");
      exit (EXIT_FAILURE);
   }

/*-- READ RAW socket --*/
   if((*sd_recv = socket(AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0) {
      perror ("opening raw read socket");
      exit (EXIT_FAILURE);
   }
   if((rsflags = fcntl(*sd_recv, F_GETFL)) == -1) {
      perror ("fcntl(sd_recv,F_GETFL)");
      exit (EXIT_FAILURE);
   }
   if (fcntl(*sd_recv, F_SETFL, rsflags | O_NONBLOCK) == -1) {
      perror ("fcntl(sd_recv,F_SETFL)");
      exit (EXIT_FAILURE);
   }
}

void synscan(int sd_send, int sd_recv, int *pschkd, hdrinfo targetinfo) {
   struct tcphdr tcp;
   int i, j;
   long timeout;
   
   for(j = 0; j < SERVICES; j++)
      pschkd[j] = PORT_CLOSED;
   
   for(i = 0; i < SERVICES; i++) {
      targetinfo.dport = service[i].port;

      targetinfo.seq = rand();
      sendtcp(sd_send, &targetinfo, SYN);
      
      usleep(10);
      
      if(gettcp(sd_recv, ip(targetinfo.dest), &tcp)) {
	for(j = 0; j < SERVICES; j++)
           if(service[j].port == ntohs(tcp.th_sport)) {
              if(ntohs(tcp.th_win)) {
                 pschkd[j] = 0;
	         break;
              }
	   }
      }
   }
   
   timeout = time(NULL) + TIMEOUT;

   while (timeout > time(NULL)) {
      if(gettcp(sd_recv, ip(targetinfo.dest), &tcp)) {
	for(j = 0; j < SERVICES; j++)
           if(service[j].port == ntohs(tcp.th_sport)) {
              if(ntohs(tcp.th_win)) {
                 pschkd[j] = 0;
	         break;
              }
	   }
      }
   }
}

void finscan(int sd_send, int sd_recv, int *pschkd, hdrinfo targetinfo) {
   struct tcphdr tcp;
   int i, j;
   long timeout;
   u_short tcpflags = 0;
   
   for(i = 0; i < SERVICES; i++) {
      targetinfo.dport = service[i].port;

      targetinfo.seq = rand();
         
      switch (i % 3) { /* this will help prevent scan detection */
         case 0: tcpflags = 0x0; break; /* NULL */
         case 1: tcpflags = FIN; break; /* FIN  */
         case 2: tcpflags = PSH; break; /* PUSH */
      }
      sendtcp(sd_send, &targetinfo, tcpflags);
      
      usleep(10);
      
      if(gettcp(sd_recv, ip(targetinfo.dest), &tcp)) {
	for(j = 0; j < SERVICES; j++)
           if(service[j].port == ntohs(tcp.th_sport)) {
              pschkd[j] = PORT_CLOSED;
	      break;
	   }
      }
   }
   
   timeout = time(NULL) + TIMEOUT;

   while (timeout > time(NULL)) {
      if(gettcp(sd_recv, ip(targetinfo.dest), &tcp)) {
	for(j = 0; j < SERVICES; j++)
	   if(service[j].port == ntohs(tcp.th_sport)) {
	      pschkd[j] = PORT_CLOSED;
              break;
	   }
      }
   }
}

