
#include "mns.h"
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <rpc/pmap_clnt.h>

int rpcscan (struct in_addr, const char *);
struct node *rpc (struct in_addr, struct node *);

const static char *rpcservices[7] =
{"mountd", "rstatd", "nlockmgr", "status", "nist", "ttdb", "autofsd"};

struct node *
rpc (struct in_addr hostip, struct node *head)
{

  int i = 0;
  struct node *walk = head;

  /*lazy style */
  stralloc (walk, "\n   RPC ->");

  for (i = 0; i < 7; i++)
    if ((rpcscan (hostip, rpcservices[i]) == 1))
      stralloc (walk, rpcservices[i]);

  if ((head->next)->next == NULL)
    {
      free (head->string);
      free (head->next);
      bzero (head, sizeof (struct node));
      return head;
    }
  else
    return walk;
}

int 
rpcscan (struct in_addr hostip, const char *program)
{
  SAin server_addr;
  struct pmaplist *head = NULL;
  int sockett = RPC_ANYSOCK;
  struct timeval minutetimeout;
  register CLIENT *client;
  struct rpcent *rpc;

  server_addr.sin_addr = hostip;
  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons (PMAPPORT);
  minutetimeout.tv_sec = TIMEOUT << 1;
  minutetimeout.tv_usec = 0;

  signal (SIGALRM, &sighandler);
  /* cause clnttcp_create uses connect() */
  alarm (TIMEOUT << 1);

  if ((client = clnttcp_create (&server_addr, PMAPPROG,
				PMAPVERS, &sockett, 50, 500)) == NULL)
    {
      alarm (0);
      return 0;
    }
  alarm (0);

  if (clnt_call (client, PMAPPROC_DUMP, (xdrproc_t) xdr_void, NULL,
	     (xdrproc_t) xdr_pmaplist, &head, minutetimeout) != RPC_SUCCESS)
    return 0;

  if (head != NULL)
    for (; head != NULL; head = head->pml_next)
      if ((rpc = getrpcbynumber (head->pml_map.pm_prog)))
	if (strcmp (rpc->r_name, program) == 0)
	  return (1);

  return 0;
}
